/*
** Module   : TempTableHelper.java
** Abstract : Helper class which generates temp table DDL
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description----------------------------
** 001 ECF 20080311  @37509  Created initial version. Extracted classes
**                           from DatabaseManager. Added computed column
**                           support. Added capability of making index
**                           names unique by session.
** 002 CA  20080424  @38106  Refactored to allow generation of global temp
**                           tables and indices.
** 003 CA  20080430  @38159  Refactored to allow correct generation for
**                           indices: the indices must be retrieved from
**                           the "main" database or the dmo_index.xml file
**                           depending on the database type: dirty or not.
** 004 ECF 20080512  @38505  Refactored to allow access from sub-package.
**                           Made class and several static methods public.
** 005 ECF 20080611  @38699  Removed queryIndexData(). The common, static
**                           variant of this method in Persistence is now
**                           used instead.
** 006 CA  20080815  @39469  Support API change in DBUtils.
** 007 ECF 20081107  @40406  Moved multiplex ID position in index. It used
**                           to be the first element in any index, but
**                           that sometimes resulted in a less efficient
**                           index being selected for a query plan. Now it
**                           is either the last element, or second to last
**                           (if the primary key is part of the index).
** 008 SVL 20120331          Upgraded to Hibernate 4.
** 009 SVL 20140106          Changed access modifier for get() function.
** 010 OM  20140410          Fixed support case-sensitive fields in indexes.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import org.apache.commons.logging.*;
import org.hibernate.*;
import org.hibernate.cfg.*;
import org.hibernate.dialect.*;
import org.hibernate.engine.spi.*;
import org.hibernate.id.factory.*;
import org.hibernate.mapping.*;
import org.hibernate.type.*;
import com.goldencode.p2j.persist.dialect.*;

/**
 * Abstract helper class which generates and caches SQL phrases to create and 
 * drop temporary tables, and to create indexes in those tables.  This class
 * contains the following service methods intended for use by external
 * classes:
 * <ul>
 *   <li>{@link #sqlTempTableCreate} - gets a dialect-specific SQL statement
 *       to create a particular temp table.
 *   <li>{@link #sqlTempTableDrop} - gets a dialect-specific SQL statement
 *       to drop a particular temp table.
 *   <li>{@link #sqlTempTableIndexCreate} - gets a list of dialect-specific
 *       SQL statements to create indexes for a particular temp table.
 *   <li>{@link #sqlTempTableIndexDrop} - gets a list of dialect-specific
 *       SQL statements to drop indexes for a particular temp table.
 * </ul>
 */
public abstract class TempTableHelper
{
   /** Logger */
   protected static final Log LOG = LogFactory.getLog(TempTableHelper.class);
   
   /** Key which will be replaced by a unique suffix in certain DDL */
   static final String UNIQUE_SUFFIX_KEY = "1";
   
   /** Length of "create table" prefix in SQL create table statements */
   private static final int CREATE_PREFIX_LEN = "create table".length();   
   
   /** Map of DMO interfaces to helper objects */
   private static final Map<Class<?>, TempTableHelper> cache =
      new HashMap<Class<?>, TempTableHelper>();
   
   /** Name of table schema */
   protected final String schema;

   /** The used database */
   protected final Database database;

   /** Object used by Hibernate for create table and index statements */
   private final TableMapping defaultTableMapping;
   
   /** Dialect-specific SQL statement(s) to create temp table(s) */
   private final List<String> sqlCreateTables;
   
   /** Dialect-specific SQL statement(s) to drop temp table(s) */
   private final List<String> sqlDropTables;
   
   /** Dialect-specific SQL statement(s) to create temp table index(es) */
   private final List<String> sqlCreateIndexes;
   
   /** Dialect-specific SQL statement(s) to drop temp table index(es) */
   private final List<String> sqlDropIndexes;
   
   /** DMO interface which provides the public API to the temp table */
   private final Class<?> dmoIface;
   
   /**
    * Return the SQL statement text used to create a temporary table for the
    * given DMO class, using the specified database dialect.
    *
    * @param   database
    *          The <code>dmoIface</code>'s backing database.
    * @param   dmoIface
    *          DMO interface which represents the temp table.
    *
    * @return  SQL statement(s) for the temp table create operation(s), using
    *          syntax specific to the dialect of the given Persistence
    *          instance.
    *
    * @throws  RuntimeException
    *          if any error occurs generating the SQL statement.  This will
    *          generally indicate a configuration problem, rather than a
    *          condition which should be handled by application code.
    */
   public static Iterator<String> sqlTempTableCreate(Database database,
                                                     Class<?> dmoIface)
   {
      return get(database, dmoIface).sqlCreateTables.iterator();
   }
   
   /**
    * Return the SQL statement text used to drop a temporary table for the
    * given DMO class, using the specified database dialect.
    *
    * @param   database
    *          The <code>dmoIface</code>'s backing database.
    * @param   dmoIface
    *          DMO interface which represents the temp table.
    *
    * @return  SQL statement(s) for the table drop operation, using syntax
    *          specific to the given dialect.
    */
   public static Iterator<String> sqlTempTableDrop(Database database,
                                                   Class<?> dmoIface)
   {
      return get(database, dmoIface).sqlDropTables.iterator();
   }
   
   /**
    * Return a list of SQL statements used to create the indexes for a
    * temporary table for the given DMO class, using the specified database
    * dialect.  The list may be empty.
    *
    * @param   database
    *          The <code>dmoIface</code>'s backing database.
    * @param   dmoIface
    *          DMO interface which represents the temp table.
    *
    * @return  SQL statements for the temp table index create operations,
    *          using syntax specific to the given dialect.
    *
    * @throws  RuntimeException
    *          if any error occurs generating the SQL statements.  This will
    *          generally indicate a configuration problem, rather than a
    *          condition which should be handled by application code.
    */
   public static Iterator<String> sqlTempTableIndexCreate(Database database,
                                                          Class<?> dmoIface)
   {
      return get(database, dmoIface).sqlCreateIndexes.iterator();
   }
   
   /**
    * Return a list of SQL statements used to drop the indexes for a
    * temporary table for the given DMO class, using the specified database
    * dialect.  The list may be empty.
    *
    * @param   database
    *          The <code>dmoIface</code>'s backing database.
    * @param   dmoIface
    *          DMO interface which represents the temp table.
    *
    * @return  SQL statements for the temp table index create operations,
    *          using syntax specific to the given dialect.
    *
    * @throws  RuntimeException
    *          if any error occurs generating the SQL statements.  This will
    *          generally indicate a configuration problem, rather than a
    *          condition which should be handled by application code.
    */
   public static Iterator<String> sqlTempTableIndexDrop(Database database,
                                                        Class<?> dmoIface)
   {
      return get(database, dmoIface).sqlDropIndexes.iterator();
   }
   
   /**
    * Get the backing temp table name associated with the given DMO
    * interface, primarily for debugging and error reporting purposes.
    * 
    * @param   database
    *          The <code>dmoIface</code>'s backing database.
    * @param   dmoIface
    *          DMO interface for which a table name is needed.
    *
    * @return  Name of the temp table associated with the given interface.
    */
   static String getTableName(Database database, Class<?> dmoIface)
   {
      String schema = DatabaseManager.getSchema(database);
      Class<?> cls = DMOIndex.getImplementingClass(schema, dmoIface);
      
      return DatabaseManager.getBackingTableName(database, cls);
   }

   /**
    * Get the computed cols for a certain table.
    *
    * @param   database
    *          The <code>dmoIface</code>'s backing database.
    * @param   dmoIface
    *          DMO interface for which a table name is needed.
    *
    * @return  A map of computed index components which represent the 
    *          computed columns.
    */
   static Map<String, P2JIndexComponent> getComputedCols(Database database,   
                                                         Class<?> dmoIface)
   {
      String schema = DatabaseManager.getSchema(database);
      
      return get(database, dmoIface).getComputedCols(schema, dmoIface);
   }
      
   /**
    * Factory method for obtaining a helper.  First checks for an entry
    * in the cache, creating a new instance if necessary.
    * <p>
    * If the associated database is the {@link DatabaseManager#TEMP_TABLE_DB 
    * database}, an instance of class {@link LocalTempTableHelper} will be 
    * returned. Else, an instance of class {@link DirtyTempTableHelper} will 
    * be returned.
    *
    * @param   database
    *          The <code>dmoIface</code>'s backing database.
    * @param   dmoIface
    *          DMO interface which provides the public API to the temp
    *          table.
    *
    * @return  Helper object associated with the given DMO interface.
    */
   static TempTableHelper get(Database database,
                              Class<?> dmoIface)
   {
      synchronized (cache)
      {
         // Note:  it is safe to share a common cache between "local" temp
         // tables and "dirty" temp tables because there is no overlap of DMO
         // types between these two uses.
         
         TempTableHelper helper = cache.get(dmoIface);
         if (helper == null)
         {
            P2JDialect dialect =
               (P2JDialect) DatabaseManager.getDialect(database);
            
            if (database.equals(DatabaseManager.TEMP_TABLE_DB))
            {
               helper = new LocalTempTableHelper(dialect, dmoIface);
            }
            else
            {
               helper = new DirtyTempTableHelper(dialect, dmoIface, database);
            }
            
            // cache it only if the DMO was resolved; the only case when the
            // DMO is not resolved is when the ORMHandler.mapClass will need 
            // to get the database index metadata before the associated HBM  
            // file is mapped (in DatabaseManager.registerDatabase)
            if (helper.dmoIface != null)
            {
               cache.put(dmoIface, helper);
            }
         }
         
         return helper;
      }
   }
   
   /**
    * Constructor.
    *
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   dmoIface
    *          DMO interface which provides the public API to the temp
    *          table.
    * @param   database
    *          The database for which we generate the sql's.
    */
   TempTableHelper(P2JDialect dialect, 
                   Class<?> dmoIface,
                   Database database)
   {
      this.schema = DatabaseManager.getSchema(database);
      this.database = database;
      this.defaultTableMapping = new TableMapping(database);
      
      Class<?> dmoClass = DMOIndex.getImplementingClass(schema, dmoIface);
      Table primary = DatabaseManager.getBackingTable(database, dmoClass);
      
      // the only situation when the backing table can't be resolved is when
      // this method is called by the ORMHandler.mapClass to retrieve the
      // associated index metadata (from the dmo_index.xml file or from the
      // database)
      if (primary == null)
      {
         this.dmoIface = null;
         
         // the DMO was not registered yet, return
         sqlCreateTables = Collections.emptyList();
         sqlCreateIndexes = Collections.emptyList();
         sqlDropTables = Collections.emptyList();
         sqlDropIndexes = Collections.emptyList();
         return;
      }
      
      this.dmoIface = dmoIface;
      
      List<Table> secondaries =
         DatabaseManager.getSecondaryTables(database, primary.getName());
      
      sqlCreateTables = Collections.unmodifiableList(
         generateSQLCreateTableList(dialect, primary, secondaries));
      sqlCreateIndexes = Collections.unmodifiableList(
         generateSQLCreateIndexList(dialect, dmoIface, primary, secondaries));
      sqlDropTables = Collections.unmodifiableList(
         generateSQLDropTableList(dialect, primary, secondaries));
      sqlDropIndexes = Collections.unmodifiableList(
         generateSQLDropIndexList(dialect, dmoIface, primary, secondaries));
   }
   
   /**
    * Abstract method which will return the needed create table string:
    * "create local temporary table" or "create temporary table".
    * 
    * @return  The create table string.
    */
   protected abstract String getCreateTableStatement();
   
   /**
    * Abstract method which will return the an additional index suffix (to 
    * make the index unique across sessions), if needed.
    * 
    * @return  An additional index suffix, if needed.
    */
   protected abstract String getAdditionalIndexSuffix();
   
   /**
    * Abstract method which will return the computed columns for the required
    * DMO.
    * 
    * @param   schema
    *          The schema used for this DMO.
    * @param   dmoIface
    *          The DMO for which to retrieve the computed columns.
    * 
    * @return  A map with the computed columns associated with this DMO.
    */
   protected abstract Map<String, P2JIndexComponent> 
   getComputedCols(String schema, Class<?> dmoIface);
   
   /**
    * Abstract method which will return the indexes associated with a certain
    * DMO.
    * 
    * @param   schema
    *          The schema used for this DMO.
    * @param   dmoIface
    *          The DMO for which to retrieve the indexes.
    * 
    * @return  An iterator for the indexes associated with this DMO.
    */
   protected abstract Iterator<P2JIndex> getIndexes(String schema, 
                                                    Class<?> dmoIface);
   
   /**
    * Populate the list of dialect-specific SQL statements used to create
    * the required temp table(s).
    *
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   primary
    *          Hibernate database table configuration for primary temp
    *          table.
    * @param   secondaries
    *          Optional list of Hibernate database table configurations
    *          for secondary temp tables.  These are the tables which
    *          support temp table fields with extents, if any.  May be
    *          <code>null</code> if no such tables exist.
    *
    * @return  List of SQL statements.
    */
   private List<String> generateSQLCreateTableList(P2JDialect dialect,
                                                   Table primary,
                                                   List<Table> secondaries)
   {
      ArrayList<String> stmts = new ArrayList<String>();
      
      // Primary table first;  secondaries are dependent upon it.
      stmts.add(generateSQLCreateTable(primary, dialect));
      
      // Secondary tables, if any.
      if (secondaries != null)
      {
         Iterator<Table> iter = secondaries.iterator();
         while (iter.hasNext())
         {
            stmts.add(generateSQLCreateTable(iter.next(), dialect));
         }
      }
      
      stmts.trimToSize();
      
      if (LOG.isDebugEnabled())
      {
         StringBuilder buf =
            new StringBuilder("TEMP TABLE CREATE statements for ");
         buf.append(primary.getName());
         buf.append(":\n");
         Iterator<String> iter = stmts.iterator();
         while (iter.hasNext())
         {
            buf.append("   ");
            buf.append(iter.next());
            buf.append("\n");
         }
         LOG.debug(buf.toString());
      }
      
      return stmts;
   }
   
   /**
    * Generate the dialect-specific SQL statement used to create a primary
    * or secondary temp table.
    *
    * @param   table
    *          Hibernate database table helper.
    * @param   dialect
    *          Hibernate database dialect helper.
    *
    * @return  SQL statement text.
    *
    * @throws  RuntimeException
    *          if there is any error generating the SQL statement.
    */
   private String generateSQLCreateTable(Table table, P2JDialect dialect)
   {
      if (dialect.needsComputedColumns())
      {
         return generateComputedColumnsSQLCreate(table, dialect);
      }
      
      return generateStandardSQLCreate(table, dialect);
   }
   
   /**
    * Generate the DDL necessary to create a local, temporary table with
    * computed columns which right trim whitespace and optionally uppercase
    * text columns.
    * 
    * @param   table
    *          Hibernate table definition.
    * @param   dialect
    *          P2J database dialect.
    * 
    * @return  DDL statement.
    */
   private String generateComputedColumnsSQLCreate(Table table, P2JDialect dialect)
   {
      String sql = null;
      Dialect hDial = (Dialect) dialect;
      Column lastCol = null;
      
      Map<String, P2JIndexComponent> computedCols = getComputedCols(schema, dmoIface);
      
      StringBuilder buf = new StringBuilder(getCreateTableStatement());
      buf.append(table.getQuotedName(hDial));
      buf.append(" (");
      
      PrimaryKey pKey = table.getPrimaryKey();
      
      Iterator cols = table.getColumnIterator();
      for (int i = 0; cols.hasNext(); i++)
      {
         if (i > 0)
         {
            buf.append(", ");
         }
         
         Column col = (Column) cols.next();
         
         String qName = col.getQuotedName(hDial);
         String sqlType = col.getSqlType(hDial, defaultTableMapping);
         
         buf.append(qName);
         buf.append(" ");
         buf.append(sqlType);
         
         String rawName = col.getName();
         if (dialect.isComputedColumn(rawName))
         {
            // TODO: we assume here that all/both cc prefixes have the same length
            String ccPrefix = dialect.getComputedColumnPrefix(false);
            
            rawName = rawName.substring(ccPrefix.length());
            P2JIndexComponent idxComp = computedCols.get(rawName);
            
            // TODO:  this bit is probably dialect-specific, currently hard-coded to H2 syntax.
            if (idxComp != null)
            {
               boolean ignoreCase = idxComp.isIgnoreCase();
               
               // Note:  this assumes lastCol contains the column to which the
               // computed column refers.
               String baseName = lastCol.getQuotedName(hDial);
               
               buf.append(" as ");
               if (ignoreCase)
               {
                  buf.append("upper(");
               }
               buf.append("rtrim(");
               buf.append(baseName);
               buf.append(", ' \t\n\r')");
               if (ignoreCase)
               {
                  buf.append(")");
               }
            }
         }
         
         if (!col.isNullable())
         {
            buf.append(" not null");
         }
         
         lastCol = col;
      }
      
      // Handle primary key.
      if (pKey != null)
      {
         buf.append(", ");
         buf.append(pKey.sqlConstraintString(hDial));
      }
      
      buf.append(")");
      
      sql = buf.toString();
      
      return sql;
   }
   
   /**
    * Generate the DDL necessary to create a standard temp table (without
    * computed columns).
    * 
    * @param   table
    *          Hibernate table definition.
    * @param   dialect
    *          P2J database dialect.
    * 
    * @return  DDL statement.
    */
   private String generateStandardSQLCreate(Table table, P2JDialect dialect)
   {
      String sql = null;
      
      try
      {
         sql = table.sqlCreateString((Dialect) dialect,
                                     defaultTableMapping,
                                     null,
                                     null);
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         throw new RuntimeException(
            "Error generating create temp table SQL statement", exc);
      }
      
      if (sql == null)
      {
         throw new RuntimeException("Cannot create temp table");
      }
      
      StringBuilder buf = new StringBuilder(getCreateTableStatement());
      sql = sql.toLowerCase().substring(CREATE_PREFIX_LEN);
      buf.append(sql);
      
      return buf.toString();
   }
   
   /**
    * Populate the list of dialect-specific SQL statements used to create
    * any indexes required by the temp table(s).
    *
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   dmoIface
    *          DMO interface which provides the public API to the temp
    *          table.
    * @param   primary
    *          Hibernate database table configuration for primary temp
    *          table.
    * @param   secondaries
    *          Optional list of Hibernate database table configurations
    *          for secondary temp tables.  These are the tables which
    *          support temp table fields with extents, if any.  May be
    *          <code>null</code> if no such tables exist.
    *
    * @return  List of SQL statements.  Will be empty if no indices are
    *          defined.
    */
   private List<String> generateSQLCreateIndexList(P2JDialect dialect,
                                                   Class<?> dmoIface,
                                                   Table primary,
                                                   List<Table> secondaries)
   {
      Iterator<P2JIndex> iter = getIndexes(schema, dmoIface);
      
      return generateSQLCreateIndexList(dialect, 
                                        dmoIface, 
                                        primary, 
                                        secondaries, 
                                        iter);
   }
   
   /**
    * Populate the list of dialect-specific SQL statements used to create
    * any indexes required by the temp table(s).
    *
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   dmoIface
    *          DMO interface which provides the public API to the temp
    *          table.
    * @param   primary
    *          Hibernate database table configuration for primary temp
    *          table.
    * @param   secondaries
    *          Optional list of Hibernate database table configurations
    *          for secondary temp tables.  These are the tables which
    *          support temp table fields with extents, if any.  May be
    *          <code>null</code> if no such tables exist.
    * @param   iter
    *          The iterator for the collection which contains the needed 
    *          indexes. 
    *          
    * @return  List of SQL statements.  Will be empty if no indices are
    *          defined.
    */
   protected List<String> generateSQLCreateIndexList(P2JDialect dialect,
                                                     Class<?> dmoIface,
                                                     Table primary,
                                                     List<Table> secondaries,
                                                     Iterator<P2JIndex> iter)
   {
      ArrayList<String> stmts = new ArrayList<String>();
      
      while (iter.hasNext())
      {
         P2JIndex next = iter.next();
         stmts.add(generateSQLCreateIndex(dialect,
                                          next,
                                          primary,
                                          true,
                                          false));
      }
      
      // Handle indices for secondary tables, if any.
      /*
      if (secondaries != null)
      {
         Iterator<Table> iter2 = secondaries.iterator();
         while (iter2.hasNext())
         {
            Table table = iter2.next();
            Iterator idxIter = table.getIndexIterator();
            while (idxIter.hasNext())
            {
               Index next = (Index) idxIter.next();
               stmts.add(next.sqlCreateString((Dialect) dialect,
                                              DEFAULT_TABLE_MAPPING,
                                              null,
                                              null));
            }
         }
      }
      */
      
      stmts.trimToSize();
      
      if (LOG.isDebugEnabled())
      {
         StringBuilder buf =
            new StringBuilder("INDEX CREATE statements for ");
         buf.append(primary.getName());
         buf.append(":\n");
         Iterator<String> iter3 = stmts.iterator();
         while (iter3.hasNext())
         {
            buf.append("   ");
            buf.append(iter3.next());
            buf.append("\n");
         }
         LOG.debug(buf.toString());
      }
      
      return stmts;
   }
   
   /**
    * Generate the dialect-specific SQL statement used to create an index
    * for the temp table.
    *
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   index
    *          P2J index descriptor.
    * @param   table
    *          Hibernate database table helper.
    * @param   mpid
    *          <code>true</code> to make multiplex ID the leading index
    *          component;  <code>false</code> to omit it. The multiplex ID 
    *          column will not be indexed if the {@link #schema} is not the
    *          {@link DatabaseManager#TEMP_TABLE_SCHEMA}.
    * @param   overridePK
    *          If <code>true</code>, do not add primary key column to a
    *          non-unique index;  if <code>false</code>, add primary key
    *          column to a non-unique index.
    *
    * @return  DDL statement text.
    */
   protected String generateSQLCreateIndex(P2JDialect dialect,
                                           P2JIndex index,
                                           Table table,
                                           boolean mpid,
                                           boolean overridePK)
   {
      List<Column> columns = new ArrayList<Column>();
      
      // Extract index information into a Hibernate-friendly form.
      LinkedHashSet<String> names = index.getNameSet(dialect);
      Iterator<String> iter = names.iterator();
      while (iter.hasNext())
      {
         columns.add(new Column(iter.next()));
      }
      
      // Multiplex ID is a component of every index, unless explicitly
      // requested to omit.
      if (mpid && schema.equals(DatabaseManager.TEMP_TABLE_SCHEMA))
      {
         columns.add(new Column(TemporaryBuffer.MULTIPLEX_FIELD_NAME));
      }
      
      // Add primary key to non-unique indexes to permit deterministic,
      // fine-grained sorting.
      boolean unique = index.isUnique();
      if (!unique && !overridePK)
      {
         columns.add(new Column("id"));
      }
      
      // Possibly make the index name globally unique so databases which do
      // not support the creation of local indexes do not fail.
      StringBuilder buf = new StringBuilder(index.getName());
      buf.append(getAdditionalIndexSuffix());
      String name = buf.toString();
      
      // Compose the string.
      return Index.buildSqlCreateIndexString((Dialect) dialect,
                                             name,
                                             table,
                                             columns.iterator(),
                                             unique,
                                             null,
                                             null);
   }
   
   /**
    * Populate the list of dialect-specific SQL statements used to drop
    * any indexes required by the temp table(s).
    *
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   dmoIface
    *          DMO interface which provides the public API to the temp
    *          table.
    * @param   primary
    *          Hibernate database table configuration for primary temp
    *          table.
    * @param   secondaries
    *          Optional list of Hibernate database table configurations
    *          for secondary temp tables.  These are the tables which
    *          support temp table fields with extents, if any.  May be
    *          <code>null</code> if no such tables exist.
    *
    * @return  List of SQL statements.  Will be empty if no indices are
    *          defined.
    */
   private List<String> generateSQLDropIndexList(P2JDialect dialect,
                                                 Class<?> dmoIface,
                                                 Table primary,
                                                 List<Table> secondaries)
   {
      Iterator<P2JIndex> iter = getIndexes(schema, dmoIface);
      
      return generateSQLDropIndexList(dialect, 
                                      dmoIface, 
                                      primary, 
                                      secondaries, 
                                      iter); 
   }
   
   /**
    * Populate the list of dialect-specific SQL statements used to drop
    * any indexes required by the temp table(s).
    *
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   dmoIface
    *          DMO interface which provides the public API to the temp
    *          table.
    * @param   primary
    *          Hibernate database table configuration for primary temp
    *          table.
    * @param   secondaries
    *          Optional list of Hibernate database table configurations
    *          for secondary temp tables.  These are the tables which
    *          support temp table fields with extents, if any.  May be
    *          <code>null</code> if no such tables exist.
    * @param   iter
    *          The iterator for the collection which contains the needed 
    *          indexes.
    *          
    * @return  List of SQL statements.  Will be empty if no indices are
    *          defined.
    */
   protected List<String> generateSQLDropIndexList(P2JDialect dialect,
                                                   Class<?> dmoIface,
                                                   Table primary,
                                                   List<Table> secondaries,
                                                   Iterator<P2JIndex> iter)
   {
      ArrayList<String> stmts = new ArrayList<String>();
      
      // Handle indices for secondary tables, if any.
      /*
      if (secondaries != null)
      {
         Iterator<Table> iter2 = secondaries.iterator();
         while (iter2.hasNext())
         {
            Table table = iter2.next();
            Iterator idxIter = table.getIndexIterator();
            while (idxIter.hasNext())
            {
               Index next = (Index) idxIter.next();
               stmts.add(generateSQLDropIndex(dialect, next.getName()));
            }
         }
      }
      */
      
      // Handle indices for primary table.
      while (iter.hasNext())
      {
         P2JIndex next = iter.next();
         stmts.add(generateSQLDropIndex(dialect, next.getName()));
      }
      
      stmts.trimToSize();
      
      if (LOG.isDebugEnabled())
      {
         StringBuilder buf =
            new StringBuilder("INDEX DROP statements for ");
         buf.append(primary.getName());
         buf.append(":\n");
         Iterator<String> iter3 = stmts.iterator();
         while (iter3.hasNext())
         {
            buf.append("   ");
            buf.append(iter3.next());
            buf.append("\n");
         }
         LOG.debug(buf.toString());
      }
      
      return stmts;
   }
   
   /**
    * Generate the dialect-specific SQL statement used to drop an index for
    * the temp table.
    *
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   name
    *          Index base name.
    *
    * @return  SQL statement text.
    */
   protected String generateSQLDropIndex(P2JDialect dialect, String name)
   {
      StringBuilder buf = new StringBuilder("drop index ");
      buf.append(name);
      buf.append(getAdditionalIndexSuffix());
      
      return buf.toString();
   }
   
   /**
    * Populate the list of dialect-specific SQL statements used to drop
    * the specified temp table(s).  Secondary tables, if any, are dropped
    * before the primary table, since they will have foreign key relations
    * to the primary table.
    *
    * @param   dialect
    *          Hibernate database dialect helper.
    * @param   primary
    *          Hibernate database table configuration for primary temp
    *          table.
    * @param   secondaries
    *          Optional list of Hibernate database table configurations
    *          for secondary temp tables.  These are the tables which
    *          support temp table fields with extents, if any.  May be
    *          <code>null</code> if no such tables exist.
    *
    * @return  List of SQL statements.
    */
   protected List<String> generateSQLDropTableList(P2JDialect dialect,
                                                   Table primary,
                                                   List<Table> secondaries)
   {
      ArrayList<String> stmts = new ArrayList<String>();
      
      // Secondary tables, if any, must be deleted before primary table.
      if (secondaries != null)
      {
         Iterator<Table> iter = secondaries.iterator();
         while (iter.hasNext())
         {
            stmts.add(generateSQLDropTable(iter.next(), dialect));
         }
      }
      
      // Primary table is deleted last.
      stmts.add(generateSQLDropTable(primary, dialect));
      
      stmts.trimToSize();
      
      if (LOG.isDebugEnabled())
      {
         StringBuilder buf =
            new StringBuilder("TEMP TABLE DROP statements for ");
         buf.append(primary.getName());
         buf.append(":\n");
         Iterator<String> iter = stmts.iterator();
         while (iter.hasNext())
         {
            buf.append("   ");
            buf.append(iter.next());
            buf.append("\n");
         }
         LOG.debug(buf.toString());
      }
      
      return stmts;
   }
   
   /**
    * Generate the dialect-specific SQL statement used to drop a temp
    * table.
    *
    * @param   table
    *          Hibernate database table helper.
    * @param   dialect
    *          Hibernate database dialect helper.
    *
    * @return  SQL statement text.
    *
    * @throws  RuntimeException
    *          if there is any error generating the SQL statement.
    */
   private String generateSQLDropTable(Table table, P2JDialect dialect)
   {
      String sql = table.sqlDropString((Dialect) dialect, null, null);
      if (sql == null)
      {
         throw new RuntimeException("Cannot drop table");
      }
      
      return sql;
   }
   
   /**
    * Implementation of Hibernate's internal <code>Mapping</code> interface,
    * required when generating dialect-specific, SQL create temp table
    * statements.
    * <p>
    * <b>Portability Note:</b>  Although the Hibernate interface
    * <code>org.hibernate.engine.Mapping</code> is <code>public</code>, it
    * appears to be intended primarily for Hibernate internal use.  Thus, we
    * have a dependency here on Hibernate 3.0's internal implementation.
    * Should this interface be removed or the API which uses it change in
    * future versions, the mechanism we use to generate dialect-specific,
    * SQL for temp table create and drop functionality will need to change.
    */
   private static class TableMapping
   implements Mapping
   {
      /** Database associated with this table mapping */
      private final Database database;
      
      /**
       * Constructor.
       *
       * @param   database
       *          Database associated with this table mapping.
       */
      TableMapping(Database database)
      {
         this.database = database;
      }
      
      /**
       * Get the Hibernate type for the specified DMO's primary key identifier
       * property.
       *
       * @param   dmoCls
       *          Fully qualified DMO implementation class name.
       *
       * @return  Type of identifier.
       *
       * @throws  MappingException
       *          if the class name is not recognized within the scope of the
       *          database associated with this table mapping.
       */
      public Type getIdentifierType(String dmoCls)
      throws MappingException
      {
         return getPersistentClass(dmoCls).getIdentifier().getType();
      }
      
      /**
       * Get the property name for the specified DMO's primary key identifier.
       *
       * @param   dmoCls
       *          Fully qualified DMO implementation class name.
       *
       * @return  Identifier property name.
       *
       * @throws  MappingException
       *          if the class name is not recognized within the scope of the
       *          database associated with this table mapping.
       */
      public String getIdentifierPropertyName(String dmoCls)
      throws MappingException
      {
         return getPersistentClass(dmoCls).getIdentifierProperty().getName();
      }

      /**
       * Allow access to the id generator factory.
       *
       * @return <code>null</code>, always.
       */
      public IdentifierGeneratorFactory getIdentifierGeneratorFactory()
      {
         return null;
      }

      /**
       * Get the type of the specified property of the specified DMO class.
       *
       * @param  dmoCls
       *         DMO class.
       * @param  propertyName
       *         The name of the property.
       *
       * @return The type of the specified property.
       *
       * @throws MappingException
       *          If DMO class or property name is invalid.
       */
      public Type getReferencedPropertyType(String dmoCls, String propertyName)
      throws MappingException
      {
         return getPersistentClass(dmoCls).getProperty(propertyName).getType();
      }

      /**
       * Get the Hibernate type for the specified DMO property.
       *
       * @param   dmoCls
       *          Fully qualified DMO implementation class name.
       * @param   property
       *          DMO property name as specified in the Hibernate mapping
       *          document for <code>dmoCls</code>.
       *
       * @return  Type of the specified property.
       *
       * @throws  MappingException
       *          if the class name is not recognized within the scope of the
       *          database associated with this table mapping;
       *          if the given property name is not recognized.
       */
      public Type getPropertyType(String dmoCls, String property)
      throws MappingException
      {
         PersistentClass pc = getPersistentClass(dmoCls);
         Property prop = pc.getProperty(property);
         if (prop == null)
         {
            throw new MappingException("Unknown property:  " +
                                       dmoCls + "." + property);
         }
         
         return prop.getType();
      }
      
      /**
       * Helper to get the Hibernate configuration object associated with the
       * specified DMO class.
       *
       * @param   dmoCls
       *          Fully qualified DMO implementation class name.
       *
       * @return  Hibernate configuration object associated with DMO class.
       *
       * @throws  MappingException
       *          if the class name is not recognized within the scope of the
       *          database associated with this table mapping.
       */
      private PersistentClass getPersistentClass(String dmoCls)
      throws MappingException
      {
         Configuration config =
            DatabaseManager.getHibernateConfiguration(database);
         if (config == null)
         {
            return null;
         }
         
         PersistentClass pc = config.getClassMapping(dmoCls);
         if (pc == null)
         {
            throw new MappingException("Unknown DMO class:  " + dmoCls);
         }
         
         return pc;
      }
   }
}
